#include "camera/camera.h"
#include "list/typed_list.h"
#include "mem_arena.h"
#include "mem_utils.h"
#include "misc/misc_utils.h"
#include "rasteriser/rasteriser.h"
#include "vector/vec.h"
#include "window/window.h"
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_keycode.h>
#include <stdlib.h>

int main(void) {
  colour_t bg =
      (colour_t){.rgba.r = 27, .rgba.g = 38, .rgba.b = 79, .rgba.a = 255};
  camera_t camera = {
      .position = {.x = 0.0f, .y = 0.0f, .z = 0.0f},
      .rotation = {.x = 0.0f, .y = 0.0f, .z = 0.0f},
  };
  vec3f_t viewport = {.x = 1.0f, .y = 1.0f, .z = 1.0f};

  window_t window = {0};

  if (!init_window(&window, 800, 800, "CG From Scratch Rasteriser")) {
    return EXIT_FAILURE;
  }

  Arena *arena = NULL;
  u64 main_arena_capacity = 20ull * 1024ull * 1024ull * 1024ull;
  if (!wapp_mem_arena_init(&arena, main_arena_capacity, WAPP_MEM_ALLOC_RESERVE,
                           false)) {
    return EXIT_FAILURE;
  }

  list_vec3f_t *vertices = list_create(vec3f_t, arena);
  if (!vertices) {
    return EXIT_FAILURE;
  }
  // Vertices
  vec3f_t verts[] = {
      (vec3f_t){1, 1, 1},    (vec3f_t){-1, 1, 1},  (vec3f_t){-1, -1, 1},
      (vec3f_t){1, -1, 1},   (vec3f_t){1, 1, -1},  (vec3f_t){-1, 1, -1},
      (vec3f_t){-1, -1, -1}, (vec3f_t){1, -1, -1},
  };
  for (u64 i = 0; i < ARR_LEN(verts); ++i) {
    list_append(vec3f_t, arena, vertices, verts[i]);
  }

  list_scene_triangle_t *triangles = list_create(scene_triangle_t, arena);
  if (!triangles) {
    return EXIT_FAILURE;
  }
  // Triangles
  scene_triangle_t tris[] = {
      (scene_triangle_t){0, 1, 2, (colour_t){.colour = 0xff0000ff}},
      (scene_triangle_t){0, 2, 3, (colour_t){.colour = 0xff0000ff}},
      (scene_triangle_t){4, 0, 3, (colour_t){.colour = 0x00ff00ff}},
      (scene_triangle_t){4, 3, 7, (colour_t){.colour = 0x00ff00ff}},
      (scene_triangle_t){5, 4, 7, (colour_t){.colour = 0x0000ffff}},
      (scene_triangle_t){5, 7, 6, (colour_t){.colour = 0x0000ffff}},
      (scene_triangle_t){1, 5, 6, (colour_t){.colour = 0xffff00ff}},
      (scene_triangle_t){1, 6, 2, (colour_t){.colour = 0xffff00ff}},
      (scene_triangle_t){4, 5, 1, (colour_t){.colour = 0xff00ffff}},
      (scene_triangle_t){4, 1, 0, (colour_t){.colour = 0xff00ffff}},
      (scene_triangle_t){2, 6, 7, (colour_t){.colour = 0x00ffffff}},
      (scene_triangle_t){2, 7, 3, (colour_t){.colour = 0x00ffffff}},
  };
  for (u64 i = 0; i < ARR_LEN(tris); ++i) {
    list_append(scene_triangle_t, arena, triangles, tris[i]);
  }

  model_t cube = {.vertices = vertices, .triangles = triangles};
  instance_t cube_01 = {
      .model = &cube,
      .transform =
          {
              .scale = 1.0f,
              .rotation = (vec3f_t){0},
              .position = (vec3f_t){-1.5f, 0.0f, 8.0f},
          },
  };
  instance_t cube_02 = {
      .model = &cube,
      .transform =
          {
              .scale = 1.0f,
              .rotation = {0},
              .position = (vec3f_t){1.7f, 1.5f, 7.0f},
          },
  };

  mat3x4f_t proj_matrix = {
      (vec4f_t){viewport.z * window.width / viewport.x, 0.0f, 0.0f, 0.0f},
      (vec4f_t){0.0f, viewport.z * window.height / viewport.y, 0.0f, 0.0f},
      (vec4f_t){0.0f, 0.0f, 1.0f, 0.0f},
  };
  mat4x4f_t cam_inverse_rotation = {0};
  mat4x4f_t cam_inverse_translation = {0};
  mat4x4f_t cam_matrix = {0};
  mat3x4f_t proj_cam_matrix = {0};

  bool running = true;
  SDL_Event event = {0};

  // i32 w_min = ((i32)window.half_width) * -1;
  // i32 w_max = (i32)window.half_width;
  // i32 h_min = ((i32)window.half_height) * -1;
  // i32 h_max = (i32)window.half_height;

  Arena *frame_arena = NULL;
  u64 frame_arena_capacity = 10ull * 1024ull * 1024ull * 1024ull;
  if (!wapp_mem_arena_init(&frame_arena, frame_arena_capacity,
                           WAPP_MEM_ALLOC_RESERVE, false)) {
    return EXIT_FAILURE;
  }

  while (running) {
    while (SDL_PollEvent(&event)) {
      switch (event.type) {
      case SDL_QUIT:
        running = false;
        break;
      }
    }

    cam_inverse_rotation = inverse_camera_rotation_matrix(&camera);
    cam_inverse_translation = inverse_camera_translation_matrix(&camera);
    cam_matrix = mul_mat4x4f(cam_inverse_rotation, cam_inverse_translation);
    proj_cam_matrix = mul_mat3x4f_by_mat4x4f(proj_matrix, cam_matrix);

    clear_window(&window, bg);

    render_instance(&window, arena, proj_cam_matrix, &cube_01);
    render_instance(&window, arena, proj_cam_matrix, &cube_02);

    swap_buffers(&window);

    wapp_mem_arena_clear(frame_arena);
  }

  wapp_mem_arena_destroy(&frame_arena);
  wapp_mem_arena_destroy(&arena);
  close_window(&window);

  return EXIT_SUCCESS;
}
